#!/usr/bin/perl

# ISABS backend script (RPM version);
# Usage: isabs_rpm.pl [--action] <IPVL> [<ACF> ...]

################### sub routines (version compatition) ###################

# a >  b == -1
# a == b == 0  
# a <  b == +1

sub sign {
    my ($value) = @_;

    return 1 if ($value > 0);
    return 0 if ($value == 0);
    return -1 if ($value < 0);
}

sub compare_literal {
    my ($comp_a, $comp_b) = @_;
    my (%literal_version, $weight_a, $weight_b);

    %literal_version = (
	"alpha" => -9,
	"beta" => -8,
	"gamma" => -7,
	"delta" => -6,

	"pre" => -3,
	"prerelease" => -3,
	"preview" => -3,

	"test" => -2,

	"rc" => -1,
	"releasecandidate" => -1,

	"release" => 0,
	"final" => 0,

	"stable" => +1,
	"patchlevel" => +1,
	"pl" => +1,
    );

    $weight_a = 0;
    $weight_b = 0;

    if ($comp_a =~ /\D+/) {
	$comp_a =~ tr/[A-Z]/[a-z]/;
	if (exists($literal_version{$comp_a})) { 
	    $weight_a = $literal_version{$comp_a};
	} else {
	    $weight_a = -10000;
	}
    }
    if ($comp_b =~ /\D+/) {
	$comp_b =~ tr/[A-Z]/[a-z]/;
	if (exists($literal_version{$comp_b})) {
	    $weight_b = $literal_version{$comp_b};
	} else {
	    $weight_b = -10000;
	}
    }

    if ($comp_a ne "" && $comp_b eq "") {
	if ($comp_a =~ /\d+/) {
	    return -1;
	} elsif ($weight_a == -10000) {
	    return -1;
	} else {
	    return sign($weight_a * -1);
	}
    }

    if ($comp_a eq "" && $comp_b ne "") {
	if ($comp_b =~ /\d+/) {
	    return +1;
	} elsif ($weight_b == -10000) {
	    return +1;
	} else {
	    return sign($weight_b);
	}
    }

    if ($comp_a =~ /\d+/) {
	if ($comp_b =~ /\d+/) {
	    if ($comp_a > $comp_b) {
		return -1;
	    } elsif ($comp_a < $comp_b) {
		return +1;
	    }
	} else {
	    # Is it right?
	    return +1;
	}
    }

    if ($weight_a == -10000) {
	if ($weight_b == -10000) {
	    # both a and b are not literal_version (such as simple alphabet).
	    if ($comp_a gt $comp_b) {
		return -1;
	    } elsif ($comp_a lt $comp_b) {
		return +1;
	    } else {
		return 0;
	    }
	} else {
	    # a isn't literal_version, but b is.
	    return -1;
	}
    } elsif ($weight_b == -10000) {
	# a is literal_version, but b isn't.
	return +1;
    } else {
	# both a and b are literal_version.
	if ($weight_a > $weight_b) {
	    return -1;
	} elsif ($weight_a < $weight_b) {
	    return +1;
	} else {
	    return 0;
	}
    }

    return 0;
}

sub compare_version {
    my ($a, $b) = @_;
    my (@dim_a, @dim_b, $long, $proc_a, $proc_b, $comp_a, $comp_b, $i, $ret); 

    $ret = 0;

    @dim_a = split(/\./, $a);
    @dim_b = split(/\./, $b);

    if (@dim_a < @dim_b) {
	$long = @dim_b;
    } else {
	$long = @dim_a;
    }

    for ($i = 0; $i < $long; $i++) {
	if ($i > @dim_a) {
	    return +1;
	} elsif ($i > @dim_b) {
	    return -1;
	}

	$proc_a = $dim_a[$i];
	$proc_b = $dim_b[$i];

	while ($proc_a ne "" || $proc_b ne "") {
	    $comp_a = "";
	    $comp_b = "";
	    if ($proc_a =~ /^(\D+)(.*)$/) {
		$comp_a = $1;
		$proc_a = $2;
	    } elsif ($proc_a =~ /^(\d+)(.*)$/) {
		$comp_a = $1;
		$proc_a = $2;
	    }
	    if ($proc_b =~ /^(\D+)(.*)$/) {
		$comp_b = $1;
		$proc_b = $2;
	    } elsif ($proc_b =~ /^(\d+)(.*)$/) {
		$comp_b = $1;
		$proc_b = $2;
	    }

	    if ($comp_a =~ /\D+/ && $comp_b eq "") {
		# a is literal, no b.
		# matrix
		# a == last, b == last : return comparison result directly.
		# a != last, b == last : return comparison result directly.
		# a == last, b != last : return +1
		# a != last, b != last : return +1
		return +1 if (@dim_b - 1 > $i);
		return compare_literal($comp_a, $comp_b);
	    } elsif ($comp_b =~ /\D+/ && $comp_a eq "") {
		# b is literal, no a.
		return -1 if (@dim_a - 1 > $i);
		return compare_literal($comp_a, $comp_b);
	    }

	    $ret = compare_literal($comp_a, $comp_b);
	    return $ret if ($ret != 0);
	};
    }
    return 0;
}

sub compare_version_release {
    my ($a, $b) = @_;
    my ($ver_a, $ver_b, $rev_a, $rev_b, $ret);

    ($ver_a, $rev_a) = split(/-/, $a);
    ($ver_b, $rev_b) = split(/-/, $b);

    $ret = compare_version($ver_a, $ver_b);
    return $ret if ($ret != 0);

    return compare_version($rev_a, $rev_b);
}

################### sub routines (file parsing) ###################

sub check_action
{
    my ($action) = @_;

    return $action if ($action =~ /(ce|pu|ss|none)/);
    if ($action =~ /(pf|packetfiltering)/) {
	$need_packet_filtering = 1;
	return "pf";
    }
    return "ce" if ($action eq "commandexecute");
    return "ss" if ($action eq "serviceshutdown");
    return "pu" if ($action eq "packageupgrade");
    return "none" if ($action eq "nothing");

    return "";
}

sub parse_acf
{
    my ($package, $version, $action, $option, $i, $checked_action);

    for ($i = 1; $i < @ARGV; $i++) {
	open (FD_ACF, $ARGV[$i]) || do { print "can\'t open Action Configuration File $ARGV[$i]!"; next; };
	while (<FD_ACF>) {
	    next if (/^(\s*\#|\s*$)/);

	    /^\s*(\S+):??(\S*)\s+(\S+)\s+(.*)$/;
	    ($package, $version, $action, $option) = ($1, $2, $3, $4);

	    $checked_action = check_action($action);
	    if ($checked_action eq "") {
		print "illegal action $action for package $package, ignored.";
		next;
	    }
	    $acf_action{$package} = $checked_action;
	    $acf_version{$package} = $version if ($version ne "");
	    $acf_option{$package} = $option if ($option ne "");
	}
	close(FD_ACF);
    }
}

sub check_installed_version
{
    my ($packageS, $versionS, $installed_version, $i);
    my ($version_range_comment, $version_range, $comment, $version_from, $version_to);

    open (FD_IPVL, $ARGV[0]) || die "can't open IPVL file";
    while (<FD_IPVL>) {
	chomp;

	($packageS, $versionS, $comment) = split (/\s+/);

	foreach $package (split(/\|/, $packageS)) {

	    if (exists($acf_version{$package})) {
		$installed_version = $acf_version{$package};
	    } else {
                ### RPM version ###
		open(FD_PMS, "/bin/rpm -q --qf='%{VERSION}-%{RELEASE}' $package |") || die "can't execute rpm";
		$installed_version = <FD_PMS>;

		### dpkg version (commented out)
		#open(FD_PMS, "/usr/bin/dpkg -l $package | tail +6 | awk '{ print \$3 }' |") || die "can't execute rpm";
		#$installed_version = <FD_PMS>;
	    }

	    close(FD_PMS);
	    chomp($installed_version);

	    next if ($installed_version =~ /^package .* installed$/);
	    next if ($installed_version eq "");
	    next if ($installed_version eq "<none>");  ### for dpkg

	    #foreach $version_range_comment (split(/\,/, $versionS)) {
	    foreach $version_range (split(/\,/, $versionS)) {
		#($version_range, $comment) = split(/\#/, $version_range_comment);

		if ($version_range ne "") {
		    if ($version_range =~ /~/ && $version_range ne "") {
		        ($version_from, $version_to) = split(/~/, $version_range);
		        next if (compare_version_release($version_to, $installed_version) > 0);
		        next if (compare_version_release($version_from, $installed_version) < 0);
		    } else {
		        next if ($installed_version ne $version_range);
		    }
		}

		$insecure_package_version{$package} = $installed_version;
		$insecure_package_version_from{$package} = $version_from;
		$insecure_package_version_to{$package} = $version_to;
		$comment =~ tr/\|/,/;
		$insecure_package_comment{$package} = $comment if ($comment ne "");
		last;
	    }
	}
    }
    close(FD_IPVL);
}

################### sub routines (port/protocol checking) ###################

sub get_package_by_cmdline
{
    my ($pid, $packageS) = @_;

    open(FD_SCRIPT, "/proc/$pid/cmdline") || next;
    $c = <FD_SCRIPT>;
    close(FD_SCRIPT);

    $packageS = "";
    foreach $script (split(/\000/, $c)) {
	open(FD_RPM, "/bin/rpm -qf --qf='%{NAME}' $script |") || die "can't execute rpm";
	$package = <FD_RPM>;
	close(FD_RPM);

	chomp($package);
	next if ($package =~ /^file .* package$/);  # not found.

	if ($packageS eq "") {
	    $packageS = $package;
	} else {
	    $packageS .= ",$package";
	}
    }
    return $packageS;
}

sub get_package_list_by_xinetd
{
    my ($xinetdconf, $service, $disable, $protocol, $port, $package, $dummy);

    opendir(DD_XINETD, "/etc/xinetd.d") || die "can't open /etc/xinetd.d/";
    while ($xinetdconf = readdir(DD_XINETD)) {
	open(FD, "/etc/xinetd.d/$xinetdconf") || next;
	while (<FD>) {
	    if (/^\s*service\s+(\w+)/) {
		$service = $1;
	    } elsif (/^\s*disable\s*=\s*(\w+)/) {
		$disable = $1;
	    } elsif (/^\s*protocol\s*=\s*(\w+)/) {
		$protocol = $1;
	    }
	}
	close(FD);
	$disable =~ tr/[A-Z]/[a-z]/;
	if ($disable eq "no") {
	    $port = (getservbyname($service, $protocol))[2];
	    open(FD_RPM, "/bin/rpm -qf --qf='%{NAME}' /etc/xinetd.d/$xinetdconf |") || die "can't execute rpm";
	    $package = <FD_RPM>;
	    chomp($package);
	    close(FD_RPM);
	    next if ($package =~ /^file .* package/);
	    $xinetd_package_by_port{$port} = $package;
	    if (exists($port_protocol_by_package{$package})) {
	        $port_protocol_by_package{$package} .= "$port/$protocol";
	    } else {
	        $port_protocol_by_package{$package} = "$port/$protocol";
	    }
	}
    }
    closedir(DD_XINETD);
}

sub get_package_by_xinetd
{
    my ($port) = @_;

    return $xinetd_package_by_port{$port} if (exists($xinetd_package_by_port{$port}));
    return "";
}

sub get_socket_infomation 
{
    my ($protocol, @socketS_dim, $socket, $command, $package);

    opendir(DD_PROC, "/proc") || die "can't open /proc directory";
    while ($i = readdir(DD_PROC)) {
	next unless ($i =~ /^\d+$/);
	opendir(DD_PID, "/proc/$i/fd") || next;
	$command = "";
	while ($k = readdir(DD_PID)) {
	    $l = readlink("/proc/$i/fd/$k");
	    if ($l =~ /^socket:\[(\d+)\]$/) {
		$socket = $1;
		if ($command eq "") {
		    $command = readlink("/proc/$i/exe") || next;
		    open(FD_RPM, "/bin/rpm -qf --qf='%{NAME}' $command |") || die "can't execute rpm";
		    $package = <FD_RPM>;
		    close(FD_RPM);
		}
		$command_by_socket{$socket} = $command;
		$pid_by_socket{$socket} = $i;
		if ($command =~ m%(/bin/sh|/bin/bash|/bin/tcsh|/bin/csh|/usr/bin/perl|/usr/bin/python)%) {
		    $package_by_socket{$socket} = get_package_by_cmdline($i);
		} elsif ($command eq "/usr/sbin/xinetd") {
		    $package_by_socket{$socket} = get_package_by_xinetd($port_by_socket{$socket});
		} else {
		    $package_by_socket{$socket} = $package;
		}
	    }
	}
	closedir(DD_PID);
    }
    closedir(DD_PROC);

    foreach $protocol ("tcp", "udp", "raw") {
	open(FD, "/proc/net/$protocol") || do { print "can\'t open TCP info!"; next; };
	<FD>;
	while(<FD>)  {
	    @socketS_dim = split(/[\s:]+/);
	    $socket = $socketS_dim[14];
	    $address_by_socket{$socket} = $socketS_dim[2];  # perhaps not used.
	    $port_by_socket{$socket} = hex("0x$socketS_dim[3]");
	    $protocol_by_socket{$socket} = $protocol;
	}
	close(FD);
    }
}

sub get_port_protocol_by_package
{
    local (%xinetd_package_by_port, %port_by_socket, %address_by_socket, %protocol_by_socket, %package_by_socket, %pid_by_socket);
    my ($socket, $package);

    get_package_list_by_xinetd;
    get_socket_infomation;

    foreach $socket (keys(%package_by_socket)) {
	next unless (exists($port_by_socket{$socket}));
	foreach $package (split(/,/, $package_by_socket{$socket})) {
	    if (exists($port_protocol_by_package{$package})) {
		$port_protocol_by_package{$package} .= ",$port_by_socket{$socket}/$protocol_by_socket{$socket}";
	    } else {
		$port_protocol_by_package{$package} = "$port_by_socket{$socket}/$protocol_by_socket{$socket}";
	    }
	}
    }
}

################### sub routines (actions) ###################

sub execute
{
    my ($commandline) = @_;

    if ($need_action == 1) {
	open (FD_SH, "| $shell") || do { print "can't execute shell $shell, ignored"; return; };
	print FD_SH $commandline;
	close (FD_SH);
    print "done: ";
    }
    print "$commandline\n";
}

# Packet Filtering
# Option allowed; port/protocol[!exceptaddress][^networkinterface]
# comma (,) to separate multiple option.
# pipe (|)  to separate multiple entry in same factor.
sub do_packet_filtering
{
    my ($package) = @_;
    my ($parse, $protocol, $port, $ipaddress_network, $network_interface);
    my ($protocolS, $portS, $ipaddress_networkS, $network_interfaceS);

    if (exists($acf_option{$package})) {
	foreach $parse (split(/[,\s]+/,$acf_option{$package})) {
	    if ($parse =~ /^([\w\|]+)\/([\w\|]+)@??([!\w\.\|\/]+)\^??([!\w]+)/) {
		($portS, $protocolS, $ipaddress_netmaskS, $network_interfaceS) = ($1, $2, $3, $4);
#	    } elsif ($parse =~ /^([\w\|]+)\/([\w\|]+)/) {
#		($portS, $protocolS) = ($1, $2);
	    } else {
		print "illegal packet filtering option $acf_option{$package}, skipping";
		next;
	    }
	    foreach $port (split(/\|/, $portS)) {
		foreach $protocol (split(/\|/, $protocolS)) {
		    if ($network_interfaceS ne "") {
			foreach $network_interface (split/\|/, $network_interfaceS) {
			    $network_interface = "! $1" if ($network_interface =~ /^\!(\w+)$/);
			    if ($ipaddress_netmaskS ne "") {
				foreach $ipaddress_netmask (split(/\|/, $ipaddress_netmaskS)) {
				    $ipaddress_netmask = "! $1" if ($ipaddress_netmask =~ /^\!(.+)$/);
				    execute("/sbin/iptables -A ISABS -p $protocol --dport $port -s $ipaddress_netmask -i $network_interface -j DROP");
				}
			    } else {
				execute("/sbin/iptables -A ISABS -p $protocol --dport $port -i $network_interface -j DROP");
			    }
			}
		    } else {
			if ($ipaddress_netmaskS ne "") {
			    foreach $ipaddress_netmask (split(/\|/, $ipaddress_netmaskS)) {
				$ipaddress_netmask = "! $1" if ($ipaddress_netmask =~ /^!(.*)$/);
				execute("/sbin/iptables -A ISABS -p $protocol --dport $port -s $ipaddress_netmask -j DROP");
			    }
			} else {
			    execute("/sbin/iptables -A ISABS -p $protocol --dport $port -j DROP");
			}
		    }			
		}
	    }
 	}
    } else {
	foreach $parse (split(/[,\s]+/, $port_protocol_by_package{$package})) {
	    ($port, $protocol) = split(/\//, $parse);
	    execute("/sbin/iptables -A ISABS -p $protocol --dport $port -j DROP");
	}
    }
}

# Command Execute
# Option needed; command line to execute.
sub do_command_execute
{
    my ($package) = @_;

    if ($acf_option{$package} eq "") {
	print "The action for $package is ce(command execute), but no arguments. So I do packet filtering instead of it";
	return do_packet_filtering($package);
    } else {
	execute($acf_option{$package});
    }
}


# Service Shutdown
# No option. Ignored if there is.
sub do_service_shutdown
{
    my ($package) = @_;
    my ($service);

    open(FD_PMS, "/bin/rpm -ql $package |") || die "can't execute rpm";
    while (<FD_PMS>) {
	chomp;
	if (m%^/etc/init.d/% || m%^/etc/rc.d/init.d/%) {
	    execute("$_ stop") if (-x "$_");
	} elsif (m%/etc/xinetd.d/(\w+)%) {
	    execute("/sbin/chkconfig $1 off");
	    $need_xinetd_reload = 1;
	}
    }
    close(FD_PMS);
}

sub do_action
{
    my ($package);

    foreach $package (keys(%insecure_package_version)) {
	if (exists($acf_action{$package})) {
	    $action = $acf_action{$package};
	} else {
	    $action = $default_action;
	}

	if ($action eq "pf") {
	    do_packet_filtering($package);
	}  if ($action eq "ce") {
	    do_command_execute($package);
	} elsif ($action eq "ss") {
	    do_service_shutdown($package);
	} elsif ($action eq "pu") {
	    # not implemented yet... perhaps impossible.
	    # do_package_update($package);
	    do_packet_filtering($package);
	}
    }
}

################### sub routines (reports) ###################

sub report_problem
{
    my ($package);

    print "Insecure Packages\n";
    print "package: insecure version (currently installed one) : comment\n";
    print "=============================================================\n";
    foreach $package (keys(%insecure_package_version)) {
	if (exists($insecure_package_comment{$package})) {
	    print "$package: $insecure_package_version_from{$package} =< ($insecure_package_version{$package}) =< $insecure_package_version_to{$package} : see $insecure_package_comment{$package}\n";
	} else {
	    print "$package: $insecure_package_version_from{$package} =< ($insecure_package_version{$package}) =< $insecure_package_version_to{$package}\n";
        }
    }
}

################### main routines ###################

sub main
{

    local %acf_version, %acf_action, %acf_option;
    
    local %insecure_package_version;
    local %insecure_package_version_from;
    local %insecure_package_version_to;

    local %insecure_package_comment;
    local %port_protocol_by_package;
    local $default_action = "pf";
    local $need_packet_filtering = 0;
    local $need_xinetd_reload = 0;
    local $need_action = 0;
    local $shell = $ENV{"SHELL"};

    $need_packet_filtering = 1 if ($default_action eq "pf");

    if ($ARGV[0] eq "--action") {
	shift(@ARGV);
	$need_action = 1;
    }
    die "usage: isabs_rpm.pl [--action] <IPVL> [<ACF> ...]" if (@ARGV == 0);

    parse_acf;

    check_installed_version;

    exit 0 if (keys(%insecure_package_version) == 0);

    report_problem;

    get_port_protocol_by_package;

    if ($need_packet_filtering == 1) {
	print "\n========================================\n";
	print "DO ACTION MODE running.\n";
	print "Executed commands are below:\n";
	print "========================================\n";
    } else {
	print "\n========================================\n";
	print "NO ACTION MODE running.\n";
	print "Commands to do are below:\n";
	print "========================================\n";
    }
    execute "/sbin/iptables -E ISABS ISABS_OLD";
    execute "/sbin/iptables -N ISABS";
    execute "/sbin/iptables -I INPUT -j ISABS";
    do_action;
    execute "/sbin/iptables -D INPUT -j ISABS_OLD";
    execute "/sbin/iptables -F ISABS_OLD";
    execute "/sbin/iptables -X ISABS_OLD";
    execute "/etc/init.d/xinetd reload" if ($need_xinetd_reload == 1);

    exit 0;
}

main;

