#!/usr/bin/perl

require LWP::UserAgent;
require File::Temp;
use File::Temp ();
#use Crypt::GPG;
use POSIX;

sub check_ipvl
{
    my ($ipvlfile) = @_;
    my $temporary_file = new File::Temp( UNLINK => 0,
					 SUFFIX => '.ipvl2' );

    $ENV{"LANG"} = "C";

    my $verify =
	system ("/usr/bin/gpg","-q","--yes","-o","$temporary_file","--verify","$ipvlfile");
    if ($verify == -1) {
	print "can't execute gpg";
	unlink ($temporary_file);
	return "";
    } elsif ($verify != 0) {
	print "IPVL $ipvlfile was signed with wrong GPG key\n";
	return "";
    }
    system("/usr/bin/gpg","-q","--yes","-o","$temporary_file","-d","$ipvlfile");
    unlink ($ipvlfile);
    return $temporary_file;
}

sub get_ipvl
{
    my $response, $plainipvl;

    my $temporary_file = new File::Temp( UNLINK => 0,
					 SUFFIX => '.ipvl' );
#    my $temporary_file_name = "$temporary_file_name";

    my $user_agent = LWP::UserAgent->new(env_proxy => 1,
					 keep_alive => 1,
					 timeout => 30,
					 );

    $user_agent->proxy(['http', 'ftp'], $proxy) if ($proxy ne "");

    my $i = 0;

    while ($i++ < $retry) {
	$response = $user_agent->get("$url",
				     ':content_file' => "$temporary_file"
				     );
	last if ($response->is_success);
    }

    close($temporary_file);

    unless ($response->is_success) {
	#unlink ($temporary_file_name) if (-f $temporary_file_name);
	unlink ($temporary_file) if (-f $temporary_file);
	die "downloading IPVL failed";
    }

#    unlink ($local_flie) if (-f $local_flie);

    if ($gpgpubkeyid ne "") {
	$plainipvl = check_ipvl($temporary_file);
        if ("$plainipvl" eq "") {
	    unlink ($temporary_file) if (-f $temporary_file);
	    die "checking GPG signature failed";
	} else {
	    $temporary_file = $plainipvl;
	}
    }

    system ("/bin/mv", "-f", $temporary_file, $local_file);
#	|| die "can't change filename from $temporary_file to $local_file";

    return 0;
}

sub parse_config_file
{
    my ($config_file) = @_;

    $retry = 0;

    open(FD, $config_file) || die "can't open $config_file";
    while (<FD>) {
	next if (/^\s*\#/);

	if (/^\s*(ActionConfigFile|ACF)[:\s]+([^\s\#]+)/i) {
	    $acf = $2;
	} elsif (/^\s*url[:\s]+([^\s\#]+)/i) {
	    $url = $1;
	} elsif (/^\s*localfile[:\s]+([^\s\#]+)/i) {
	    $local_file = $1
	} elsif (/^\s*proxy[:\s]+([^\s\#]+)/i) {
	    $proxy = $1;
	} elsif (/^\s*gpgpubkeyid[:\s]+([^\s\#]+)/i) {
	    $gpgpubkeyid = $1;
	} elsif (/^\s*retry[:\s]+([^\s\#]+)/i) {
	    $retry = $1;
	} elsif (/^\s*isabs[:\s]+([^\s\#]+)/i) {
	    $isabs = $1;
	}

    }
    close(FD);

    die "no URL found in config file" if ($url eq "");
    die "no LocalFile in config file" if ($local_file eq "");
    $retry = 1 if ($retry < 1);
    $isabs = "/usr/local/lib/ISABS/isabs_rpm.pl" if ($isabs eq "");

    return 0;
}

sub main
{
    local ($acf, $report_to, $proxy, $url, $local_flie, $gpgpubkeyid, $retry, $isabs);
    my $shell = $ENV{"SHELL"};

    parse_config_file($ARGV[0]);

    get_ipvl;

    exec($isabs, "--action", $local_file, $acf) || die "can't execute $isabs";
}

die "usage: isabs_frontend_pull.pl <configfile>" if (@ARGV != 1); 

main;
    
